﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using CognitiveGeology.Components.CognitiveControls.Core;
using CognitiveGeology.Components.CognitiveControls.Graph.Services;
using CognitiveGeology.Components.CognitiveControls.Model;
using Moq;
using NUnit.Framework;

namespace CognitiveGeology.CognitiveControls.Test.Graph.Services
{
    [TestFixture]
    public class LineInterpolatorTests
    {
        private const int MinimumValue = 10;
        private const int MaximumValue = 50;
        private const int StepValue = 5;

        private readonly Mock<IMathProviderFactory> mockMathProviderFactory;
        private readonly Mock<ISplineInterpolator> mockSplineInterpolator;
        private readonly Func<double, double> mockInterpolationFunction;
        private readonly DataPointCollection pointData;
        private readonly LineInterpolator underTest;

        private DataPointCollection ResultData;
        private double[] measureValues;
        private double[] propertyValues;

        public LineInterpolatorTests()
        {
            mockSplineInterpolator = new Mock<ISplineInterpolator>();
            mockInterpolationFunction = d => 2*d;
            mockSplineInterpolator.Setup(x => x.Interpolate(It.IsAny<double>())).Returns(mockInterpolationFunction);
            mockMathProviderFactory = new Mock<IMathProviderFactory>();
            mockMathProviderFactory.Setup(x => x.GetCubicSpline(It.IsAny<IEnumerable<double>>(), It.IsAny<IEnumerable<double>>()))
                .Returns(mockSplineInterpolator.Object).Callback((IEnumerable<double> a, IEnumerable<double> b) =>
                    {
                        measureValues = a.ToArray();
                        propertyValues = b.ToArray();
                    });

            pointData = new DataPointCollection()
                {
                    new DataPoint() {MeasureValue = 10, PropertyValue = 11.6},
                    new DataPoint() {MeasureValue = 20, PropertyValue = 12.5},
                    new DataPoint() {MeasureValue = 30, PropertyValue = 15.3},
                    new DataPoint() {MeasureValue = 40, PropertyValue = 15.6},
                    new DataPoint() {MeasureValue = 50, PropertyValue = 15.8},
                };

            this.underTest = new LineInterpolator(mockMathProviderFactory.Object);
        }
        
        [TestFixtureSetUp]
        public void TestFixtureSetUp()
        {
            ResultData = this.underTest.InterpolateCubic(pointData, MinimumValue, MaximumValue, StepValue);
        }

        [Test]
        public void MathProviderFactory_is_called_to_get_Cubic_Spline()
        {
            mockMathProviderFactory.Verify(x => x.GetCubicSpline(It.IsAny<IEnumerable<double>>(), It.IsAny<IEnumerable<double>>()));
        }

        [Test]
        public void MathProviderFactory_is_called_with_PointData_measureValues()
        {
            Assert.AreEqual(measureValues.Count(), pointData.Count);
            for (int index = 0; index < pointData.Count; index++)
            {
                Assert.AreEqual(pointData[index].MeasureValue, measureValues[index]);
            }
        }

        [Test]
        public void MathProviderFactory_is_called_with_PointData_propertyValues()
        {
            Assert.AreEqual(propertyValues.Count(), pointData.Count);
            for (int index = 0; index < pointData.Count; index++)
            {
                Assert.AreEqual(pointData[index].PropertyValue, propertyValues[index]);
            }
        }

        [Test]
        public void Result_contains_expected_number_of_points()
        {
            int expectedLength = ((MaximumValue - MinimumValue) / StepValue) + 1;
            Assert.AreEqual(expectedLength, ResultData.Count);
        }

        [Test] 
        public void Result_contains_expected_minimum_value()
        {
            Assert.AreEqual(MinimumValue, ResultData[0].MeasureValue);
        }

        [Test] 
        public void Result_contains_expected_maximum_value()
        {
            Assert.AreEqual(MaximumValue, ResultData.Last().MeasureValue);
        }
        
        [Test]
        public void Result_contains_interpolated_data_for_each_Measure_Value_in_supplied_range()
        {
            var index = 0;
            for (int i = MinimumValue; i < MaximumValue; i += StepValue)
            {
                Assert.AreEqual(i, ResultData[index].MeasureValue);
                Assert.AreEqual(mockInterpolationFunction(i), ResultData[index].PropertyValue);
                index++;
            }
        }
    }
}
