﻿using System;
using System.Linq;
using System.Windows.Controls;
using CognitiveGeology.Components.CognitiveControls.Graph;
using CognitiveGeology.Components.CognitiveControls.Graph.Model;
using CognitiveGeology.Components.CognitiveControls.Graph.Services;
using Moq;
using NUnit.Framework;

namespace CognitiveGeology.CognitiveControls.Test
{
    static class SplineGraphControlTests
    {
        public class SplineGraphControlTestBase
        {
            protected Mock<IPointExtractor> MockPointExtractor;
            protected Mock<ILineInterpolator> MockLineInterpolator;
            protected Mock<IDataRangeAnalyser> MockDataRangeAnalyser;
            protected SplineGraphControl UnderTest;
            protected DataPointCollection SourceData;
            protected DataPointCollection PointData;
            protected DataPointCollection LineData;
            protected RangeAnalysisResult RangeAnalysisResult; 

            protected SplineGraphControlTestBase()
            {
                this.SourceData = new DataPointCollection()
                {
                    new DataPoint() {MeasureValue = 1, PropertyValue = 1.6},
                    new DataPoint() {MeasureValue = 2, PropertyValue = 2.5},
                    new DataPoint() {MeasureValue = 3, PropertyValue = 5.3},
                    new DataPoint() {MeasureValue = 4, PropertyValue = 5.6},
                    new DataPoint() {MeasureValue = 5, PropertyValue = 5.8},
                };

                this.RangeAnalysisResult = new RangeAnalysisResult(){ Maximum = 44, Minimum = 11, StepSize = 8 };

                this.PointData = new DataPointCollection(SourceData);
                this.LineData = new DataPointCollection();

                MockPointExtractor = new Mock<IPointExtractor>();
                MockPointExtractor.Setup(x => x.ExtractPoints(It.IsAny<DataPointCollection>())).Returns(this.PointData);

                MockLineInterpolator = new Mock<ILineInterpolator>();
                MockLineInterpolator.Setup(
                    x =>
                        x.InterpolateCubic(It.IsAny<DataPointCollection>(), It.IsAny<double>(), It.IsAny<double>(),
                            It.IsAny<double>())).Returns(this.LineData);

                MockDataRangeAnalyser = new Mock<IDataRangeAnalyser>();
                MockDataRangeAnalyser.Setup(x => x.Analyse(It.IsAny<DataPointCollection>())).Returns(RangeAnalysisResult);
            }
        }

        [TestFixture]
        public class WithoutValuesSet : SplineGraphControlTestBase
        {
            public WithoutValuesSet()
            {
                UnderTest = new SplineGraphControl(MockPointExtractor.Object, MockLineInterpolator.Object, MockDataRangeAnalyser.Object);
            }

            [TestFixtureSetUp]
            public void TestFixtureSetUp()
            {
                UnderTest.Bind(SourceData);
            }

            [Test]
            public void DataSeries_is_Bound_to_SourceData()
            {
                Assert.AreSame(SourceData, UnderTest.DataSeries.ItemsSource);
            }

            [Test]
            public void PointSeries_is_Bound_to_PointData()
            {
                Assert.AreSame(PointData, UnderTest.PointSeries.ItemsSource);
            }

            [Test]
            public void PointExtractor_is_Called()
            {
                MockPointExtractor.Verify(x => x.ExtractPoints(SourceData));
            }

            [Test]
            public void LineSeries_is_Bound_to_LineData()
            {
                Assert.AreSame(LineData, UnderTest.LineSeries.ItemsSource);
            }

            [Test]
            public void LineInterpolator_is_Called_With_RangeAnalysisResult_Values()
            {
                MockLineInterpolator.Verify(x => x.InterpolateCubic(PointData, RangeAnalysisResult.Minimum, RangeAnalysisResult.Maximum, RangeAnalysisResult.StepSize));
            }
            
            [Test]
            public void DataRangeAnalyser_is_Called()
            {
                MockDataRangeAnalyser.Verify(x => x.Analyse(SourceData));
            }

            [Test]
            public void MeasureMinimumValue_is_Set_from_RangeAnalysisResult()
            {
                Assert.AreEqual(RangeAnalysisResult.Minimum, UnderTest.MeasureMinimumValue);
            }

            [Test]
            public void MeasureMaximumValue_is_Set_from_RangeAnalysisResult()
            {
                Assert.AreEqual(RangeAnalysisResult.Maximum, UnderTest.MeasureMaximumValue);
            }

            [Test]
            public void MeasureStep_is_Set_from_RangeAnalysisResult()
            {
                Assert.AreEqual(RangeAnalysisResult.StepSize, UnderTest.MeasureStep);
            }

            [Test]
            public void XAxis_has_Minimum_set_because_graph_is_horizontal()
            {
                Assert.AreEqual(UnderTest.MeasureMinimumValue, UnderTest.XAxis.MinimumValue);
            }

            [Test]
            public void XAxis_has_Maximum_set_because_graph_is_horizontal()
            {
                Assert.AreEqual(UnderTest.MeasureMaximumValue, UnderTest.XAxis.MaximumValue);
            }

            [Test]
            public void DataSeries_has_XMemberPath_set_to_MeasureValue_because_graph_is_horizontal()
            {
                Assert.AreEqual("MeasureValue", UnderTest.DataSeries.XMemberPath);
            }
            
            [Test]
            public void DataSeries_has_YMemberPath_set_to_PropertyValue_because_graph_is_horizontal()
            {
                Assert.AreEqual("PropertyValue", UnderTest.DataSeries.YMemberPath);
            }

            [Test]
            public void PointSeries_has_XMemberPath_set_to_MeasureValue_because_graph_is_horizontal()
            {
                Assert.AreEqual("MeasureValue", UnderTest.PointSeries.XMemberPath);
            }
            
            [Test]
            public void PointSeries_has_YMemberPath_set_to_PropertyValue_because_graph_is_horizontal()
            {
                Assert.AreEqual("PropertyValue", UnderTest.PointSeries.YMemberPath);
            }

            [Test]
            public void LineSeries_has_XMemberPath_set_to_MeasureValue_because_graph_is_horizontal()
            {
                Assert.AreEqual("MeasureValue", UnderTest.LineSeries.XMemberPath);
            }
            
            [Test]
            public void LineSeries_has_YMemberPath_set_to_PropertyValue_because_graph_is_horizontal()
            {
                Assert.AreEqual("PropertyValue", UnderTest.LineSeries.YMemberPath);
            }

            [Test]
            public void YAxis_is_not_inverted_because_graph_is_horizontal()
            {
                Assert.IsFalse(UnderTest.YAxis.IsInverted);
            }

            [Test]
            public void AddSplinePoint_adds_a_point_to_PointData_with_horizontal_orientation()
            {
                UnderTest.AddSplinePoint(34, 23);
                var lastPoint = UnderTest.PointData.Last();
                Assert.AreEqual(34, lastPoint.MeasureValue);
                Assert.AreEqual(23, lastPoint.PropertyValue);
            }
        }

        [TestFixture]
        public class WithValuesSet : SplineGraphControlTestBase
        {
            private const double PresetMeasureMinimumValue = 4;
            private const double PresetMeasureMaximumValue = 16;
            private const double PresetMeasureStep = 2;

            public WithValuesSet()
            {
                UnderTest = new SplineGraphControl(MockPointExtractor.Object, MockLineInterpolator.Object, MockDataRangeAnalyser.Object)
                {
                    MeasureMinimumValue = PresetMeasureMinimumValue,
                    MeasureMaximumValue = PresetMeasureMaximumValue,
                    MeasureStep = PresetMeasureStep,
                    GraphOrientation = Orientation.Vertical
                };
            }

            [TestFixtureSetUp]
            public void TestFixtureSetUp()
            {
                UnderTest.Bind(SourceData);
            }

            [Test]
            public void DataRangeAnalyser_is_not_Called()
            {
                MockDataRangeAnalyser.Verify(x => x.Analyse(SourceData), Times.Never);
            }

            [Test]
            public void MeasureMinimumValue_is_not_modified()
            {
                Assert.AreEqual(PresetMeasureMinimumValue, UnderTest.MeasureMinimumValue);
            }

            [Test]
            public void MeasureMaximumValue_is_not_modified()
            {
                Assert.AreEqual(PresetMeasureMaximumValue, UnderTest.MeasureMaximumValue);
            }

            [Test]
            public void MeasureStep_is_not_modified()
            {
                Assert.AreEqual(PresetMeasureStep, UnderTest.MeasureStep);
            }
            
            [Test]
            public void YAxis_has_Minimum_set_because_graph_is_vertical()
            {
                Assert.AreEqual(UnderTest.MeasureMinimumValue, UnderTest.YAxis.MinimumValue);
            }

            [Test]
            public void YAxis_has_Maximum_set_because_graph_is_vertical()
            {
                Assert.AreEqual(UnderTest.MeasureMaximumValue, UnderTest.YAxis.MaximumValue);
            }

            [Test]
            public void DataSeries_has_YMemberPath_set_to_MeasureValue_because_graph_is_vertical()
            {
                Assert.AreEqual("MeasureValue", UnderTest.DataSeries.YMemberPath);
            }

            [Test]
            public void DataSeries_has_XMemberPath_set_to_PropertyValue_because_graph_is_vertical()
            {
                Assert.AreEqual("PropertyValue", UnderTest.DataSeries.XMemberPath);
            }

            [Test]
            public void PointSeries_has_YMemberPath_set_to_MeasureValue_because_graph_is_vertical()
            {
                Assert.AreEqual("MeasureValue", UnderTest.PointSeries.YMemberPath);
            }

            [Test]
            public void PointSeries_has_XMemberPath_set_to_PropertyValue_because_graph_is_vertical()
            {
                Assert.AreEqual("PropertyValue", UnderTest.PointSeries.XMemberPath);
            }

            [Test]
            public void LineSeries_has_YMemberPath_set_to_MeasureValue_because_graph_is_vertical()
            {
                Assert.AreEqual("MeasureValue", UnderTest.LineSeries.YMemberPath);
            }

            [Test]
            public void LineSeries_has_XMemberPath_set_to_PropertyValue_because_graph_is_vertical()
            {
                Assert.AreEqual("PropertyValue", UnderTest.LineSeries.XMemberPath);
            }

            [Test]
            public void YAxis_is_inverted_because_graph_is_vertical()
            {
                Assert.IsTrue(UnderTest.YAxis.IsInverted);
            }

            [Test]
            public void AddSplinePoint_adds_a_point_to_PointData_with_vertical_orientation()
            {
                UnderTest.AddSplinePoint(34, 23);
                var lastPoint = UnderTest.PointData.Last();
                Assert.AreEqual(23, lastPoint.MeasureValue);
                Assert.AreEqual(34, lastPoint.PropertyValue);
            }
        }

        [TestFixture]
        public class WhenValuesChangedAfterBind : SplineGraphControlTestBase
        {
            private const double PresetMeasureMinimumValue = 4;
            private const double PresetMeasureMaximumValue = 16;
            private const double PresetMeasureStep = 2;
            
            [SetUp]
            public void SetUp()
            {
                UnderTest = new SplineGraphControl(MockPointExtractor.Object, MockLineInterpolator.Object, MockDataRangeAnalyser.Object)
                {
                    MeasureMinimumValue = PresetMeasureMinimumValue,
                    MeasureMaximumValue = PresetMeasureMaximumValue,
                    MeasureStep = PresetMeasureStep,
                    GraphOrientation = Orientation.Horizontal
                };
                UnderTest.Bind(SourceData);
            }

            [Test]
            public void MeasureMinimumValue_is_not_updated_given_null()
            {
                UnderTest.MeasureMinimumValue = null;
                Assert.AreEqual(PresetMeasureMinimumValue, UnderTest.MeasureMinimumValue);
            }
            
            [Test]
            public void MeasureMinimumValue_is_updated_given_a_number()
            {
                UnderTest.MeasureMinimumValue = 45.9;
                Assert.AreEqual(45.9, UnderTest.MeasureMinimumValue);
            }

            [Test]
            public void Axis_MinimumValue_is_updated_given_a_number()
            {
                UnderTest.MeasureMinimumValue = 45.9;
                Assert.AreEqual(45.9, UnderTest.XAxis.MinimumValue);
            }

            [Test]
            public void LineInterpolator_is_Called_With_new_Minimum_value()
            {
                UnderTest.MeasureMinimumValue = 45.9;
                MockLineInterpolator.Verify(x => x.InterpolateCubic(PointData, 45.9, PresetMeasureMaximumValue, PresetMeasureStep));
            }
            
            [Test]
            public void MeasureMaximumValue_is_not_updated_given_null()
            {
                UnderTest.MeasureMaximumValue = null;
                Assert.AreEqual(PresetMeasureMaximumValue, UnderTest.MeasureMaximumValue);
            }

            [Test]
            public void MeasureMaximumValue_is_updated_given_a_number()
            {
                UnderTest.MeasureMaximumValue = 328;
                Assert.AreEqual(328, UnderTest.MeasureMaximumValue);
            }

            [Test]
            public void Axis_MaximumValue_is_updated_given_a_number()
            {
                UnderTest.MeasureMaximumValue = 328;
                Assert.AreEqual(328, UnderTest.XAxis.MaximumValue);
            }

            [Test]
            public void LineInterpolator_is_Called_With_new_Maximum_value()
            {
                UnderTest.MeasureMaximumValue = 328;
                MockLineInterpolator.Verify(x => x.InterpolateCubic(PointData, PresetMeasureMinimumValue, 328, PresetMeasureStep));
            }
            [Test]
            public void MeasureStep_is_not_updated_given_null()
            {
                UnderTest.MeasureStep = null;
                Assert.AreEqual(PresetMeasureStep, UnderTest.MeasureStep);
            }

            [Test]
            public void MeasureStep_is_updated_given_a_number()
            {
                UnderTest.MeasureMaximumValue = 328;
                Assert.AreEqual(328, UnderTest.MeasureMaximumValue);
            }
            
            [Test]
            public void LineInterpolator_is_Called_With_new_Step_value()
            {
                UnderTest.MeasureStep = 6;
                MockLineInterpolator.Verify(x => x.InterpolateCubic(PointData, PresetMeasureMinimumValue, PresetMeasureMaximumValue, 6));
            }
        }

        [TestFixture]
        public class WhenOrientationChangedToVerticalAfterBind : SplineGraphControlTestBase
        {
            public WhenOrientationChangedToVerticalAfterBind()
            {
                UnderTest = new SplineGraphControl(MockPointExtractor.Object, MockLineInterpolator.Object, MockDataRangeAnalyser.Object)
                {
                    GraphOrientation = Orientation.Horizontal
                };
            }

            [TestFixtureSetUp]
            public void TestFixtureSetUp()
            {
                UnderTest.Bind(SourceData);
                UnderTest.GraphOrientation = Orientation.Vertical;
            }

            [Test]
            public void DataSeries_has_YMemberPath_set_to_MeasureValue_because_graph_is_vertical()
            {
                Assert.AreEqual("MeasureValue", UnderTest.DataSeries.YMemberPath);
            }

            [Test]
            public void DataSeries_has_XMemberPath_set_to_PropertyValue_because_graph_is_vertical()
            {
                Assert.AreEqual("PropertyValue", UnderTest.DataSeries.XMemberPath);
            }

            [Test]
            public void PointSeries_has_YMemberPath_set_to_MeasureValue_because_graph_is_vertical()
            {
                Assert.AreEqual("MeasureValue", UnderTest.PointSeries.YMemberPath);
            }

            [Test]
            public void PointSeries_has_XMemberPath_set_to_PropertyValue_because_graph_is_vertical()
            {
                Assert.AreEqual("PropertyValue", UnderTest.PointSeries.XMemberPath);
            }

            [Test]
            public void LineSeries_has_YMemberPath_set_to_MeasureValue_because_graph_is_vertical()
            {
                Assert.AreEqual("MeasureValue", UnderTest.LineSeries.YMemberPath);
            }

            [Test]
            public void LineSeries_has_XMemberPath_set_to_PropertyValue_because_graph_is_vertical()
            {
                Assert.AreEqual("PropertyValue", UnderTest.LineSeries.XMemberPath);
            }

            [Test]
            public void YAxis_is_inverted_because_graph_is_vertical()
            {
                Assert.IsTrue(UnderTest.YAxis.IsInverted);
            }

            [Test]
            public void YAxis_has_Minimum_set_because_graph_is_vertical()
            {
                Assert.AreEqual(UnderTest.MeasureMinimumValue, UnderTest.YAxis.MinimumValue);
            }

            [Test]
            public void YAxis_has_Maximum_set_because_graph_is_vertical()
            {
                Assert.AreEqual(UnderTest.MeasureMaximumValue, UnderTest.YAxis.MaximumValue);
            }
        }
    }
}
