﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using CognitiveGeology.Components.CognitiveControls.Histogram.Services;
using Syncfusion.UI.Xaml.Charts;

namespace CognitiveGeology.Components.CognitiveControls.Histogram
{
    /// <summary>
    /// Interaction logic for HistogramControl.xaml
    /// </summary>
    public partial class HistogramControl : UserControl
    {
        //GUI constants
        private const int StatisticsLabelPositionX = 40;
        private const int StatisticsLabelPositionY = 10;
        private const int GaussianityLabelPositionX = 40;
        private const int GaussianityLabelPositionY = 40;

        private readonly IGaussianityTester gaussianityTester;
        private int numberOfBins = -1;
        private Dictionary<double, double> binDictionary;
        private Dictionary<int, double> intervalDictionary;
        private double mean;
        private double stdev;
        private double smallestValue;
        private double biggestValue;
        private double step;

        public HistogramControl()
        {
            InitializeComponent();
            gaussianityTester = new GaussianityTester();
            
        }

        public ObservableCollection<Point> DataPoints { get; set; }
        public ObservableCollection<Point> NormalDistributionPoints { get; set; }
        public bool ShowDistributionCurve { get; set; }
        public bool ShowStatistics { get; set; }
        public bool ShowGaussianity { get; set; }
        public BinningRule Binning { get; set; }

        public void Bind(IReadOnlyCollection<double> dataset)
        {
            Build(dataset);
            Normalize();
            CreateAxisLabels();
            DataPoints = CreateObservableCollection(binDictionary);
            histogramSeries.ItemsSource = DataPoints;

            SetNormalDistributionCurve();
            CreateStatisticsLabel();
            CreateGaussianityLabel(dataset);
        }

        private void Build(IReadOnlyCollection<double> dataset)
        {
            binDictionary = new Dictionary<double, double>();
            var numberOfInstances = dataset.Count;
            var orderedData = dataset.OrderBy(d => d).ToList();
            mean = dataset.Average();
            stdev = Math.Sqrt(dataset.Select(p => Math.Pow(p - mean, 2)).Average());
            smallestValue = orderedData.First();
            biggestValue = orderedData.Last();
            step = (biggestValue == smallestValue) ? 1 : (biggestValue - smallestValue) / numberOfInstances;

            if (numberOfBins <= 0) numberOfBins = EstimateNumberOfBins(numberOfInstances, stdev);
            double interval = (biggestValue - smallestValue) / numberOfBins;

            Enumerable.Range(0, numberOfBins).ToList().ForEach(i => binDictionary.Add(i, 0));

            var values = orderedData
                .Select(val => val - orderedData.Min())
                .Select(val => Math.Floor(val/(interval*1.01)))
                .ToList();

            foreach (var val in values.Where(val => !double.IsNaN(val)))
            {
                binDictionary[val]++;
            }
            
            intervalDictionary = new Dictionary<int, double>();
            for (int i = 0; i < numberOfBins; i += 1)
            {
                intervalDictionary.Add(i, (smallestValue + (i + 1) * interval));
            }
        }

        private int EstimateNumberOfBins(double numberOfInstances, double standardDeviation)
        {
            int result;
            switch (Binning)
            {
                case BinningRule.SturgesRule:
                    {
                        result = (int)Math.Ceiling(Math.Log(numberOfInstances, 2) + 1);
                        break;
                    }
                default:
                    {
                        result = (int)Math.Ceiling(Math.Sqrt(numberOfInstances));
                        break;
                    }
            }
            return result;
        }

        private void Normalize()
        {
            double sum = binDictionary.ToList().Select(val => val.Value).Sum();
            foreach (var bin in binDictionary.ToList()) binDictionary[bin.Key] /= sum;
        }

        private void SetNormalDistributionCurve()
        {
            if (!ShowDistributionCurve)
            {
                normalSeries.ItemsSource = null;
                normalSeriesBackground.ItemsSource = null;
                return;
            }
            
            if (NormalDistributionPoints != null)
            {
                NormalDistributionPoints.Clear();
            }
            else
            {
                NormalDistributionPoints = new ObservableCollection<Point>();
            }

            var points = new List<Point>();
            for (int i = 0; i < numberOfBins; i++)
            {
                double upperBound = intervalDictionary[i];
                double lowerBound = upperBound - step;
                double midpoint = (upperBound + lowerBound) / 2;
                double x = midpoint;
                double y = GetNormalDistributionPdf(mean, stdev, x);
                points.Add(new Point(upperBound, y));
            }
            double sum = points.Select(val => val.Y).Sum();
            points.ForEach(point => NormalDistributionPoints.Add(new Point(point.X, point.Y / sum)));
            
            normalSeries.ItemsSource = NormalDistributionPoints;
            normalSeriesBackground.ItemsSource = NormalDistributionPoints;
        }

        private static double GetNormalDistributionPdf(double mean, double stdev, double x)
        {
            return ((Math.Exp(-Math.Pow(x - mean, 2) / (2 * Math.Pow(stdev, 2))) / (stdev * Math.Sqrt(2 * Math.PI))));
        }

        private void CreateAxisLabels()
        {
            var labelsAxis = intervalDictionary.OrderBy(val => val.Key).Select(val => Math.Round(val.Value, 3)).ToList();
            
            var position = 0;
            labelsAxis.ForEach(labelString =>
            {
                var cal = new ChartAxisLabel();
                cal.LabelContent = labelString;
                cal.Position = position;
                xAxis.CustomLabels.Add(cal);
                position++;
            });
        }

        private void CreateStatisticsLabel()
        {
            if (!ShowStatistics)
            {
                standardDeviationLabel.Visibility = Visibility.Collapsed;
                return;
            }

            standardDeviationLabel.X1 = StatisticsLabelPositionX;
            standardDeviationLabel.Y1 = StatisticsLabelPositionY;
            standardDeviationLabel.Text = "μ: " + Math.Round(mean, 3) + "\nσ: " + Math.Round(stdev, 3);
            standardDeviationLabel.Visibility = Visibility.Visible;
        }

        public void CreateGaussianityLabel(IEnumerable<double> dataset)
        {
            if (!ShowGaussianity)
            {
                gaussianityLabel.Visibility = Visibility.Collapsed;
                return;
            }

            var gaussianity = gaussianityTester.ComputeGaussianity(dataset.ToArray());
            
            //gaussianityLabel.X1 = GaussianityLabelPositionX;
            //gaussianityLabel.Y1 = GaussianityLabelPositionY;
            //gaussianityLabel.Text = gaussianity.IsSignificant ? "Not Gaussian" : "Gaussian";
            var color = gaussianity.IsSignificant ? Colors.Red : Colors.Green;
            gaussianityLabel.Foreground = new SolidColorBrush(color);
        }

        private static ObservableCollection<Point> CreateObservableCollection(Dictionary<double, double> data)
        {
            var result = new ObservableCollection<Point>();
            data.ToList().ForEach(val => result.Add(new Point(val.Key, val.Value)));
            return result;
        }
    }

    public enum BinningRule { SquareRootRule, SturgesRule }
}